"""Plugins Module.

This Module provides the interface for extending the functionality
of Markdown Meta Extension by adding custom CallableHandler objects.

Common CallableHandler objects are included and imported here.

Author:
    Martin Schorfmann
Since:
    2020-01-26
Version:
    2020-02-25
"""

import markdown_meta_extension.plugins.file_system_plugin
import markdown_meta_extension.plugins.package_plugin
import markdown_meta_extension.plugins.module_plugin
import markdown_meta_extension.plugins.jinja_template_plugin
from .spec import CallableWrapper, CallableHandler

# TODO: Implement call behavior for lists (i.e. for-each)



"""Jinja Template Plugin.

This plugin module provides the handlers for handling jinja2 templates
and jinja2 environments containing templates.

Author:
    Martin Schorfmann
Since:
    2020-01-28
Version:
    2020-02-25
"""

import pathlib
import types
from typing import Any, Dict, List, Text

import jinja2
import markupsafe

from markdown_meta_extension.plugins.spec import (
    CallableWrapper,
    CallableHandler
)


JINJA_FILE_SUFFIXES = [
    ".htm",
    ".html",
    ".md",
    ".markdown",
    ".j2",
    ".jinja",
    ".jinja2"
]


def check_single_template_file(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper's parent path is a template file.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrapper's parent path is a template file.
    """

    template_path: pathlib.Path = callable_wrapper.parent_imported_callable()

    return any(
        suffix.lower() in JINJA_FILE_SUFFIXES\
            for suffix in template_path.suffixes
    )


def import_single_template_file(
        callable_wrapper: CallableWrapper
) -> jinja2.Template:
    """Imports the callable wrapper's single template file.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        The imported template.
    """

    template_path: pathlib.Path = callable_wrapper.parent_imported_callable()

    template_loader = jinja2.FileSystemLoader(
        searchpath=str(template_path.parent)
    )

    # TODO: Import without loader or environment using jinja2.Template()

    template_environment = jinja2.Environment(
        loader=template_loader,
        autoescape=True
    )

    return template_environment.get_template(template_path.name)


def check_template(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper's parent import is a template.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrapper's parent import is a template.
    """

    template = callable_wrapper.parent_imported_callable()

    return isinstance(template, jinja2.Template)


def call_template(
        callable_wrapper: CallableWrapper,
        arguments: Dict[Text, Any]
) -> markupsafe.Markup:
    """Renders the template with the given arguments.

    Args:
        callable_wrapper:
            The callable wrapper around the template.
        arguments:
            The mapping of key word arguments for rendering the template.

    Returns:
        The rendered template.
    """

    if not isinstance(arguments, dict):
        arguments = dict()

    template: jinja2.Template = callable_wrapper.imported_callable

    return template.render(arguments)


def get_from_template(
        callable_wrapper: CallableWrapper,
        key: Text
) -> Any:
    """Gets the attribute from a template

    Args:
        callable_wrapper:
            The template's callable wrapper.
        key:
            The key for accessing the attribute.

    Returns:
        The accessed attribute of the template.
    """

    template: jinja2.Template = callable_wrapper.imported_callable

    if hasattr(template.module, key):  # TODO: Error handling
        return getattr(template.module, key)

    return None


def check_template_attribute(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper's parent template has the attribute.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrapper's parent template has the attribute.
    """

    attribute_name: Text = callable_wrapper.import_statement[-1]
    template: jinja2.Template = callable_wrapper.parent_imported_callable()

    # TODO: Maybe only check of import statement type or length

    return hasattr(
        template.module,
        attribute_name
    )


def import_template_attribute(
        callable_wrapper: CallableWrapper
) -> Any:
    """Imports the callable wrapper's template attribute.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        The imported template attribute.
    """

    attribute_name: Text = callable_wrapper.import_statement[-1]
    template: jinja2.Template = callable_wrapper.parent_imported_callable()

    return getattr(
        template.module,
        attribute_name
    )


def check_template_loader_directory(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper's parent path could be a template
    directory.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrapper's parent path could be a template
        directory.
    """

    directory_path: pathlib.Path = callable_wrapper.parent_imported_callable()

    for file_suffix in JINJA_FILE_SUFFIXES:

        if directory_path.glob("*" + file_suffix):
            return True

    return False


def import_template_loader_directory(
        callable_wrapper: CallableWrapper
) -> jinja2.FileSystemLoader:
    """Imports the callable wrapper's directory template loader.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        The imported directory template loader.
    """

    directory_path: pathlib.Path = callable_wrapper.parent_imported_callable()

    return jinja2.FileSystemLoader(
        searchpath=str(directory_path)
    )


def check_template_environment(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper's parent is a template environment.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrapper's parent is a template environment.
    """

    return isinstance(
        callable_wrapper.parent_imported_callable(),
        jinja2.BaseLoader
    )


def import_template_environment(
        callable_wrapper: CallableWrapper
) -> jinja2.Environment:
    """Imports the callable wrapper's template environment using
    an existing loader.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        The imported template environment.
    """

    loader: jinja2.BaseLoader = callable_wrapper.parent_imported_callable()

    template_environment = jinja2.Environment(
        loader=loader,
        autoescape=True
    )

    return template_environment


def check_template_from_environment(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper's parent template environment
    can be used to get the template.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrapper's parent template environment
        can be used to get the template.
    """

    environment: jinja2.Environment = callable_wrapper.parent_imported_callable()

    # TODO: Find better solution
    if len(callable_wrapper.import_statement) < 2:
        return False

    template_path = pathlib.Path(callable_wrapper.import_statement[1])

    return bool(environment.get_template(str(template_path)))


def import_template_from_environment(
        callable_wrapper: CallableWrapper
) -> jinja2.Environment:
    """Imports the callable wrapper's environment template

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        The imported template.
    """

    environment: jinja2.Environment = callable_wrapper.parent_imported_callable()

    # TODO: Find better solution
    template_path = pathlib.Path(callable_wrapper.import_statement[1])

    return environment.get_template(str(template_path))


JINJA_TEMPLATE_FILE_HANDLER = CallableHandler(
    name="jinja_template_file",
    requires=["file_system_file"],
    check_function=check_single_template_file,
    import_function=import_single_template_file,
    call_function=None,
    get_function=None
)

JINJA_LOADER_DIRECTORY_HANDLER = CallableHandler(
    name="jinja_loader_directory",
    requires=["file_system_directory"],
    check_function=check_template_loader_directory,
    import_function=import_template_loader_directory,
    call_function=None,
    get_function=None
)

JINJA_ENVIRONMENT_HANDLER = CallableHandler(
    name="jinja_environment",
    requires=["jinja_loader_directory"],
    check_function=check_template_environment,
    import_function=import_template_environment,
    call_function=None,
    get_function=None
)

JINJA_TEMPLATE_ENVIRONMENT_HANDLER = CallableHandler(
    name="jinja_template_environment",
    requires=["jinja_environment"],
    check_function=check_template_from_environment,
    import_function=import_template_from_environment,
    call_function=None,
    get_function=None
)

JINJA_TEMPLATE_HANDLER = CallableHandler(
    name="jinja_template",
    requires=["jinja_template_file", "jinja_template_environment"],  # TODO: Further requires
    check_function=check_template,
    import_function=None,
    call_function=call_template,
    get_function=get_from_template # TODO: Use getattr(template.module, "macro")
)

JINJA_TEMPLATE_ATTRIBUTE_HANDLER = CallableHandler(
    name="jinja_template_attribute",
    requires=["jinja_template"],
    check_function=check_template_attribute,
    import_function=import_template_attribute,
    call_function=None,
    get_function=None
)


def get_callable_handler() -> List[CallableHandler]:
    """Returns the jinja2 templates related callable handlers."""

    return [
        JINJA_TEMPLATE_FILE_HANDLER,
        JINJA_LOADER_DIRECTORY_HANDLER,
        JINJA_ENVIRONMENT_HANDLER,
        JINJA_TEMPLATE_ENVIRONMENT_HANDLER,
        JINJA_TEMPLATE_HANDLER,
        JINJA_TEMPLATE_ATTRIBUTE_HANDLER,
    ]



"""Module Handlers Plugin.

This plugin module provides the handlers for handling already imported Python
modules and using their contents like functions, classes or objects.

Author:
    Martin Schorfmann
Since:
    2020-01-28
Version:
    2020-02-25
"""

import inspect
import types
from typing import Any, List

from markdown_meta_extension.plugins.spec import (
    CallableWrapper,
    CallableHandler
)


ATTRIBUTE_SEPARATOR = "."


def check_module(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrappers current import is a Python module.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrappers current import is a Python module.
    """

    imported_callable = callable_wrapper.parent_imported_callable()

    return isinstance(imported_callable, types.ModuleType)


def check_type(
        callable_wrapper: CallableWrapper,
        type_checker: type
) -> bool:
    """Checks whether the callable wrappers current import is of the given type.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.
        type_checker:
            The type to check against.

    Returns:
        Whether the callable wrappers current import is of the given type.
    """

    return isinstance(
        callable_wrapper.parent_imported_callable,
        type_checker
    )


def check_type_function(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrappers current import is a function.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrappers current import is a function.
    """

    return check_type(callable_wrapper, types.FunctionType)


def check_type_class(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrappers current import is a class.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrappers current import is a class.
    """

    return inspect.isclass(
        callable_wrapper.parent_imported_callable()
    )


def check_type_object(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrappers current import is an object.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrappers current import is an object.
    """

    return check_type(callable_wrapper, object)


def check_type_method(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrappers current import is a method.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrappers current import is a method.
    """

    return check_type(callable_wrapper, types.MethodType)


def import_module_attribute(
        callable_wrapper: CallableWrapper
) -> Any:
    """Imports the callable wrapper's module attribute

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        The imported module attribute.
    """

    attribute_paths = callable_wrapper.import_statement[1:]
    attribute_names = list()

    for attribute_path in attribute_paths:
        attribute_names = attribute_names\
            + attribute_path.split(ATTRIBUTE_SEPARATOR)

    imported_callable = callable_wrapper.parent_imported_callable()

    for attribute_name in attribute_names:
        imported_callable = getattr(
            imported_callable,
            attribute_name
        )

    return imported_callable

MODULE_ATTRIBUTE_HANDLER = CallableHandler(
    name="module_attribute",
    requires=["file_system_module", "package"],
    check_function=check_module,
    import_function=import_module_attribute,
    call_function=None,
    get_function=None
)

MODULE_FUNCTION_HANDLER = CallableHandler(
    name="function",
    requires=["module_attribute"],
    check_function=check_type_function,
    import_function=None,
    call_function=None,
    get_function=None
)

MODULE_CLASS_HANDLER = CallableHandler(
    name="class",
    requires=["module_attribute"],
    check_function=check_type_class,
    import_function=None,
    call_function=None,
    get_function=None
)

MODULE_OBJECT_HANDLER = CallableHandler(
    name="object",
    requires=["module_attribute"],
    check_function=check_type_object,
    import_function=None,
    call_function=None,
    get_function=None
)

MODULE_METHOD_HANDLER = CallableHandler(
    name="method",
    requires=["module_attribute"],
    check_function=check_type_method,
    import_function=None,
    call_function=None,
    get_function=None
)


def get_callable_handler() -> List[CallableHandler]:
    """Returns the module and module attribute related callable handlers."""

    return [
        MODULE_ATTRIBUTE_HANDLER,
        MODULE_FUNCTION_HANDLER,
        MODULE_CLASS_HANDLER,
        MODULE_OBJECT_HANDLER,
        MODULE_METHOD_HANDLER
    ]



"""File System Handlers Plugin.

This plugin module provides the handlers for handling file paths in import
statements.

Author:
    Martin Schorfmann
Since:
    2020-01-26
Version:
    2020-03-14
"""

import importlib
import pathlib
import types
from typing import List, Text

import pathvalidate

from markdown_meta_extension.plugins.spec import (
    CallableWrapper,
    CallableHandler
)


PYTHON_FILE_SUFFIX: Text = ".py"


def check_file_system_path(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks for a file system path.

    Args:
        callable_wrapper:
            The callable wrapper to get checked.

    Returns:
        Whether the checked callable wrapper has a valid and existing path.
    """

    file_system_path = pathlib.Path(
        callable_wrapper.import_statement[0]
    )

    try:
        pathvalidate.validate_filepath(file_system_path)
    except pathvalidate.ValidationError:
        return False

    if not file_system_path.exists():
        return False

    return True


def check_file_system_file(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper refers to a file.

    Args:
        callable_wrapper:
            The callable wrapper to get checked.

    Returns:
        Whether the checked callable wrapper refers to a file.
    """

    file_system_path = callable_wrapper.parent_imported_callable()

    if not file_system_path:
        return False

    return file_system_path.is_file()


def check_file_system_directory(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper refers to a directory.

    Args:
        callable_wrapper:
            The callable wrapper to get checked.

    Returns:
        Whether the checked callable wrapper refers to a directory.
    """

    file_system_path = callable_wrapper.parent_imported_callable()

    if not file_system_path:
        return False

    return file_system_path.is_dir()


def check_file_system_module(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper refers to a python file module.

    Args:
        callable_wrapper:
            The callable wrapper to get checked.

    Returns:
        Whether the checked callable wrapper refers to a python file module.
    """

    file_path: pathlib.Path = callable_wrapper.parent_imported_callable()

    return file_path.suffix == PYTHON_FILE_SUFFIX


def import_file_system_path(
        callable_wrapper: CallableWrapper
) -> pathlib.Path:
    """Imports the callable wrapper's file system path.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        The file system path of the callable wrapper.
    """

    path = pathlib.Path(
        callable_wrapper.import_statement[0]
    )

    resolved_path = path.resolve()

    # TODO: Is Path restriction needed?
    if pathlib.Path.cwd() in resolved_path.parents:
        return path


def import_file_system_module(
        callable_wrapper: CallableWrapper
) -> types.ModuleType:
    """Imports the callable wrapper's file system Python module.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        The imported Python module.
    """

    module_path: pathlib.Path = callable_wrapper.parent_imported_callable()

    module_spec = importlib.util.spec_from_file_location(
        name=module_path.stem,
        location=module_path
    )
    imported_module = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(imported_module)

    return imported_module

FILE_SYSTEM_HANDLER = CallableHandler(
    name="file_system",
    requires=None,
    check_function=check_file_system_path,
    import_function=import_file_system_path,
)
FILE_SYSTEM_FILE_HANDLER = CallableHandler(
    name="file_system_file",
    requires="file_system",
    check_function=check_file_system_file
)
FILE_SYSTEM_DIRECTORY_HANDLER = CallableHandler(
    name="file_system_directory",
    requires="file_system",
    check_function=check_file_system_directory
)
FILE_SYSTEM_MODULE_HANDLER = CallableHandler(
    name="file_system_module",
    requires="file_system_file",
    check_function=check_file_system_module,
    import_function=import_file_system_module
)


def get_callable_handler() -> List[CallableHandler]:
    """Returns the file system related callable handler instances."""

    return [
        FILE_SYSTEM_HANDLER,
        FILE_SYSTEM_FILE_HANDLER,
        FILE_SYSTEM_DIRECTORY_HANDLER,
        FILE_SYSTEM_MODULE_HANDLER
    ]



"""Importing Module.

This module provided the extensions for Python-Markdown (markdown).

Author:
    Martin Schorfmann
Since:
    2019-12-11
Version:
    2020-03-10
"""

from pathlib import Path
import re
import string
from typing import Any, Dict, List, Text, Tuple, Union
import xml.etree.ElementTree as etree

import click
from markdown import Markdown
from markdown.extensions import Extension
from markdown.util import Registry
from markdown.blockprocessors import BlockProcessor
from markdown.inlinepatterns import Pattern, InlineProcessor
from markdown.treeprocessors import Treeprocessor
from markdown.postprocessors import Postprocessor
from markupsafe import Markup
import yaml

try:
    from pelican.readers import MarkdownReader
    from pelican.utils import pelican_open
    PELICAN = True
except ImportError:
    PELICAN = False

from markdown_meta_extension.importing import Importer, CallableWrapper


FRONTMATTER_BLOCK_PRIORITY = 210
CALL_BLOCK_PRIORITY = 200
INLINE_CALL_PRIORITY = 200
TEMPLATE_PRIORITY = 0

SYNTAX_DEFAULTS = {
    "block_start": r"{{",
    "block_end": r"}}",
    "force_display_flag": "!",
    "assignment_operator": "=",
    "attribute_operator": ".",
    "block_surrounding_tag": "div",
    "inline_surrounding_tag": "span",
    "inline_call_pattern_regex": r"(\{\{)([\s\S]+?)(\}\})"
}

class MetaExtension(Extension):
    """Meta Extension for use in Python-Markdown."""

    def extendMarkdown(self, md: Markdown):  # pylint: disable=invalid-name
        """Extension method for registering extension parts.

        Extension method for registering the additional block processors.
        Furthermore a dictionary for meta data gets initialized.

        Args:
            md:
                Markdown parser object of Python-Markdown the extensions
                get registered on.
        """

        parser_registry: Registry = md.parser.blockprocessors

        # TODO: Initialize elsewhere
        md.meta_data = dict()
        md.meta_data["variables"] = dict()
        md.meta_data["statistics"] = dict()

        # TODO: Insert Syntax Defaults

        parser_registry.register(
            FrontmatterBlockProcessor(md.parser),
            "meta-frontmatter",
            FRONTMATTER_BLOCK_PRIORITY
        )

        call_processor = CallProcessor(md.meta_data, **SYNTAX_DEFAULTS)  # TODO: Replace with kwargs

        parser_registry.register(
            CallBlockProcessor(md.parser, call_processor),
            "meta-call",
            CALL_BLOCK_PRIORITY
        )

        inline_patterns_registry: Registry = md.inlinePatterns

        inline_call_pattern_regex = self.getConfig(
            "inline_call_pattern_regex",
            SYNTAX_DEFAULTS["inline_call_pattern_regex"]
        )

        inline_patterns_registry.register(
            CallInlineProcessor(
                inline_call_pattern_regex,
                call_processor,
                md=md
            ),
            "meta-inline-call",
            INLINE_CALL_PRIORITY
        )

        post_processor_registry: Registry = md.postprocessors

        post_processor_registry.register(
            TemplatePostProcessor(md),
            "meta-template",
            TEMPLATE_PRIORITY
        )


class MetaExtensionBlockProcessor(BlockProcessor):
    """Abstract super class for Meta Extension block processors.

    Offers several common convenience methods and properties.

    Attributes:
        meta_data:
            Meta Extension data of the parser
        has_frontmatter:
            Whether the document has a frontmatter
        imports:
            Access to the Importer of meta_data.
        variables:
            Access to the variables dict of meta_data.
    """

    def get_meta_data(self) -> Dict[Text, Any]:
        """Returns the parsers meta data dict."""
        return self.parser.md.meta_data

    meta_data = property(get_meta_data)

    def get_has_frontmatter(self) -> bool:
        """Returns the has-frontmatter flag from the meta data."""
        return bool(self.meta_data.get("has-frontmatter"))

    def set_has_frontmatter(self, value: bool = True):
        """Sets the has-frontmatter flag within the meta data."""
        self.meta_data["has-frontmatter"] = value

    has_frontmatter = property(
        get_has_frontmatter,
        set_has_frontmatter
    )

    def get_imports(self) -> Importer:
        """Returns the Importer instance from the meta data."""
        return self.meta_data.get("imports")

    def set_imports(self, value: Importer):
        """Sets the Importer instance within the meta data."""
        if not isinstance(value, Importer):
            raise ValueError(Importer)

        self.meta_data["imports"] = value

    imports = property(
        get_imports,
        set_imports
    )

    def get_variables(self) -> Dict[Text, Any]:
        """Returns the variables dict from the meta data."""
        return self.meta_data.get("variables")

    def set_variables(self, value: Dict[Text, Any]):
        """Sets the variables dict within the meta data."""
        if not isinstance(value, dict):
            raise ValueError(dict)

        self.meta_data["variables"] = value

    variables = property(
        get_variables,
        set_variables
    )

    def join_lines(self, lines: List[Text]) -> Text:
        """Joins the provided lines using linebreaks and returns the string."""
        # pylint: disable=no-self-use
        return "\n".join(lines)

class FrontmatterBlockProcessor(MetaExtensionBlockProcessor):
    """Block processor for parsing frontmatter blocks.

    Attributes:
        frontmatter_delimiter_pattern:
            Regular expression to detect frontmatter block delimiters.
        has_frontmatter:
            Whether the processed document already detected a frontmatter.
    """

    frontmatter_delimiter_pattern = re.compile(r"^-{3,}\s*$")

    # TODO: Where to reset flags when processing multiple documents?

    def does_line_contain_delimiter(self, line: Text) -> bool:
        """Whether the provided line contains a frontmatter delimiter."""
        return bool(self.frontmatter_delimiter_pattern.match(line))

    def is_block_delimited(self, block: Union[Text, List[Text]]) -> bool:
        """Whether the provided text block is delimited like a frontmatter."""
        if isinstance(block, str):
            block_lines = block.splitlines()
        else:
            block_lines = block

        if len(block_lines) < 2:
            return False

        first_block_line, *_, last_block_line = block_lines

        return (
            self.does_line_contain_delimiter(first_block_line)
            and self.does_line_contain_delimiter(last_block_line)
        )

    def is_block_frontmatter(self, block: Text) -> bool:
        """Whether the provided text block is a frontmatter block."""
        block_lines = block.splitlines()

        if not block_lines:
            return False

        has_delimiters = self.is_block_delimited(block_lines)

        return has_delimiters

    def test(self, parent: etree.Element, block: Text) -> bool:
        """Tests a text block whether it is a frontmatter.

        Only the first block displaying the characteristics of frontmatters
        is considered. All following fitting blocks will return False.

        Args:
            parent:
                The parent tree node of the text block.
            block:
                The text block which gets tested.

        Returns:
            Whether the text block is a frontmatter.
        """

        if self.has_frontmatter:
            return False

        is_frontmatter = self.is_block_frontmatter(block)

        if is_frontmatter:
            self.has_frontmatter = True

        return is_frontmatter

    def run(self, parent: etree.Element, blocks: List[Text]):
        """Parses and removes the frontmatter.

        Method for parsing and removing the frontmatter, when a frontmatter
        has been found. The parsed frontmatter is used for creating an Importer
        instance, which gets attached to the parser's meta data.

        Args:
            parent:
                The parent tree node of the affected blocks.
            blocks:
                A list of text blocks to get processed.
        """

        if self.has_frontmatter:
            frontmatter_block: Text = blocks.pop(0)

            _first_line, *content_lines, _last_line = frontmatter_block.splitlines()
            frontmatter_text: Text = self.join_lines(content_lines)

            self.imports = Importer(frontmatter_text)


class CallProcessor:
    """Helper class for processing block and inline calls."""

    def __init__(self, meta_data: Dict[Text, Any], **kwargs):

        self.meta_data = meta_data

        syntax_attributes: Dict[Text, Text] = SYNTAX_DEFAULTS
        syntax_attributes.update(kwargs)

        self.block_start = syntax_attributes.get("block_start")
        self.block_end = syntax_attributes.get("block_end")

        self.force_display_flag = syntax_attributes.get("force_display_flag")
        self.assignment_operator = syntax_attributes.get("assignment_operator")
        self.attribute_operator = syntax_attributes.get("attribute_operator")

        self.block_surrounding_tag = syntax_attributes.get("block_surrounding_tag")
        self.inline_surrounding_tag = syntax_attributes.get("inline_surrounding_tag")

    def get_imports(self) -> Importer:
        """Returns the Importer instance from the meta data."""
        return self.meta_data.get("imports")

    def set_imports(self, value: Importer):
        """Sets the Importer instance within the meta data."""
        if not isinstance(value, Importer):
            raise ValueError(Importer)

        self.meta_data["imports"] = value

    imports = property(
        get_imports,
        set_imports
    )

    def get_variables(self) -> Dict[Text, Any]:
        """Returns the variables dict from the meta data."""
        return self.meta_data.get("variables")

    def set_variables(self, value: Dict[Text, Any]):
        """Sets the variables dict within the meta data."""
        if not isinstance(value, dict):
            raise ValueError(dict)

        self.meta_data["variables"] = value

    variables = property(
        get_variables,
        set_variables
    )

    def remove_block_delimiters(self, block: Text) -> Text:
        """Returns the provided block without its delimiters."""
        block = block.strip()
        return block[len(self.block_start):-len(self.block_end)]

    def has_line_force_display_flag(self, line: Text) -> bool:
        """Whether the provided line starts with a force display flag."""
        return line.startswith(self.force_display_flag)

    def remove_line_force_display_flag(self, line: Text) -> Text:
        """Returns the line without force display flag and outer whitespace."""
        _line_start, remaining_line = line.split(self.force_display_flag)

        return remaining_line.strip()

    def is_line_assignment(self, line: Text) -> bool:
        """Whether the provided line represents an assignment."""
        return self.assignment_operator in line

    def process_call(self, line: Text) -> Tuple[Text, Text]:
        """Processes a call without an assignment.

        Args:
            line:
                The line or line part to get processed

        Returns:
            A tuple consisting of the call expression part
            and the remaining line part.
            The remaining line part can be None.
        """
        # pylint: disable=no-self-use

        line = line.strip()

        if any(
                whitespace_character in line
                for whitespace_character in string.whitespace
        ):
            call_expression, remaining = line.split(maxsplit=1)
        else:
            call_expression = line
            remaining = None

        return call_expression, remaining

    def process_assignment(self, line: Text) -> Tuple[Text, Text, Text]:
        """Processes a call with an assignment.

        Args:
            line:
                The line to get processed.

        Returns:
            A tuple consisting of the parts
            variable name, call expression and what remains of the line.
            The remaining part may be None.
        """
        variable_name, remaining = line.split(
            self.assignment_operator
        )
        variable_name: Text = variable_name.strip()
        remaining: Text = remaining.strip()

        call_expression, remaining = self.process_call(remaining)

        return variable_name, call_expression, remaining

    def has_call_expression_attribute(self, call_expression: Text) -> bool:
        """Whether the provided call expression has an attribute access."""
        return self.attribute_operator in call_expression

    def join_lines(self, lines: List[Text]) -> Text:
        """Joins the provided lines using linebreaks and returns the string."""
        # pylint: disable=no-self-use
        return "\n".join(lines)

    def process_arguments(self, arguments: Union[List, Dict]) -> Union[List, Dict]:

        if isinstance(arguments, list):
            processed_arguments = [
                self.run(argument, allow_display=False) if isinstance(argument, str) and (
                    argument.startswith(self.block_start)
                    and argument.endswith(self.block_end)
                ) else argument
                for argument in arguments
            ]

        elif isinstance(arguments, dict):
            processed_arguments = {
                key: self.run(argument, allow_display=False) if isinstance(argument, str) and (
                    argument.startswith(self.block_start)
                    and argument.endswith(self.block_end)
                ) else argument
                for key, argument in arguments.items()
            }
        else:
            processed_arguments = arguments

        return processed_arguments

    def create_markup(
            self,
            markup: Any,
            is_inline: bool,
            surround: Text = None
    ) -> etree.Element:
        if markup is None:
            return None

        if hasattr(markup, "__html__"):
            markup = markup.__html__()

        try:
            markup_node = etree.fromstring(str(markup))

            if surround:
                surrounding_node = etree.Element(surround)
                surrounding_node.append(markup_node)
                markup_node = surrounding_node
        except etree.ParseError:

            if isinstance(markup, list):

                if is_inline:
                    inline_list = etree.Element(self.inline_surrounding_tag)
                    inline_list.text = ""

                    markup_list = [
                        self.create_markup(list_item, is_inline)
                        for list_item
                        in markup
                    ]

                    for index, markup_item in enumerate(markup_list):
                        if index + 1 < len(markup_list):
                            markup_item.tail = ", "

                        inline_list.append(markup_item)

                    markup_node = inline_list
                else:
                    unordered_list = etree.Element("ul")

                    for list_item in markup:
                        list_item_markup = self.create_markup(
                            list_item,
                            is_inline,
                            surround="li"
                        )
                        unordered_list.append(
                            list_item_markup
                        )

                    markup_node = unordered_list

            elif isinstance(markup, dict):
                definition_list = etree.Element("dl")

                for term, description in markup.items():
                    term_markup = self.create_markup(
                        term,
                        is_inline,
                        surround="dt"
                    )
                    description_markup = self.create_markup(
                        description,
                        is_inline,
                        surround="dd"
                    )

                    definition_list.append(
                        term_markup
                    )
                    definition_list.append(
                        description_markup
                    )

                markup_node = definition_list
            else:
                if surround:
                    surrounding_tag = surround
                else:
                    surrounding_tag = self.inline_surrounding_tag\
                        if is_inline\
                        else self.block_surrounding_tag

                markup_node = etree.Element(
                    surrounding_tag
                )
                try:
                    markup_node_child = etree.fromstring(str(markup))

                    if surround:
                        markup_node.append(markup_node_child)
                    else:
                        markup_node = markup_node_child
                except (etree.ParseError, TypeError):
                    markup_node.text = str(markup)

        return markup_node

    def run(self, block: Text, allow_display: bool = True) -> etree.Element:
        """Method which processes detected call block.

        Method which processes detected call block and returns their result.

        Args:
            block:
                The text block which gets processed.
        """
        self.meta_data["statistics"]["calls"] =\
            self.meta_data["statistics"].get("calls", 0) + 1

        # Block processing and separation
        block = self.remove_block_delimiters(block)

        try:
            variables_assignment = yaml.safe_load(block)
            if isinstance(variables_assignment, dict):
                processed_variables_assignment = self.process_arguments(
                    variables_assignment
                )
                self.variables.update(processed_variables_assignment)
                return None
            else:
                pass

        except yaml.YAMLError:
            pass

        block_lines = block.splitlines()
        first_line, *remaining_lines = block_lines

        # Force display flag detection and removal
        force_display = self.has_line_force_display_flag(first_line)

        if force_display:
            first_line = self.remove_line_force_display_flag(first_line)

        # Assignment detection and processing
        is_assignment = self.is_line_assignment(first_line)

        if is_assignment:
            # Call with assignment
            variable_name, call_expression, remaining = self.process_assignment(
                first_line
            )

            display = force_display
        else:
            # Call without assignment
            variable_name = None
            call_expression, remaining = self.process_call(first_line)
            display = True

        # Attribute access detection and processing
        has_attribute = self.has_call_expression_attribute(call_expression)

        if has_attribute:
            call_expression, *call_names = call_expression.split(
                self.attribute_operator
            )
        else:
            variable = None
            call_names = None

        # Variable or imported callable retrieval
        variable = self.variables.get(call_expression)

        if not variable:
            variable = self.imports.get(call_expression)

        callable_variable = variable

        # Optional Access of attributes to replace callable variable
        if call_names:
            for call_name in call_names:
                try:
                    callable_variable = callable_variable[call_name]
                except (KeyError, TypeError):
                    if hasattr(callable_variable, call_name):
                        callable_variable = getattr(callable_variable, call_name)
                    else:
                        try:
                            callable_variable = callable_variable[int(call_name)]
                        except ValueError:
                            callable_variable = KeyError(
                                f"Cannot find name '{call_name}' in variable."
                            )


        # Arguments parsing
        if not remaining:
            remaining = self.join_lines(remaining_lines)

        arguments = yaml.safe_load(
            remaining
        )

        arguments = self.process_arguments(arguments)

        # Abort in case of still no callable variable
        if callable_variable is None:  # FIXME: What to do?
            return None

        # Supplying the arguments to the call of the callable
        if isinstance(callable_variable, CallableWrapper):
            callable_result = callable_variable(arguments)
        else:
            if arguments:
                if isinstance(arguments, dict):
                    callable_result = callable_variable(**arguments)
                elif isinstance(arguments, list):
                    callable_result = callable_variable(*arguments)
                else:
                    callable_result = callable_variable(arguments)
            else:
                try:
                    callable_result = callable_variable()
                except TypeError:
                    callable_result = callable_variable

        # Assigning the call result in case of an assignment
        if is_assignment and variable_name:
            self.variables[variable_name] = callable_result

        # Returning the result markup node
        if allow_display and display:
            markup_output = callable_result

            is_inline = (len(remaining_lines) < 1)

            markup_node = self.create_markup(
                markup_output,
                is_inline
            )

            return markup_node
        elif not allow_display:
            return callable_result

        return None

    def __call__(self, block: Text) -> etree.Element:
        return self.run(block)


class CallBlockProcessor(MetaExtensionBlockProcessor):
    """Block processor for processing call blocks.

    Block processor which processes detected call blocks.
    Call blocks are used to invoke functions or constructor calls.

    Attributes:
        block_start:
            Regular expression of the start indicator of a call block.
        block_end:
            Regular expression of the end indicator of a call block.
        force_display_flag:
            The character used as a flag for forcing the display of the call's
            results.
        assignment_operator:
            The character used as an operator for assigning results to a
            variable.
        attribute_operator:
            The character used to access attributes or methods of an object
            variable.
    """

    def __init__(self, parser, call_processor: CallProcessor):
        super().__init__(parser)

        self.call_processor = call_processor

    def test(self, parent: etree.Element, block: Text) -> bool:
        """Method for testing a text block, whether it is a call block.

        Args:
            parent:
                The parent tree node of the text block.
            block:
                The text block to get tested, whether it is a call block.

        Returns:
            Whether the text block is a call block.
        """

        block = block.strip()

        return (
            block.startswith(self.call_processor.block_start)
            and block.endswith(self.call_processor.block_end)
        )

    def run(self, parent: etree.Element, blocks: List[Text]):
        """Method which processes detected call blocks.

        Method which processes detected call blocks and attaches their results
        to the parent tree node.

        Args:
            parent:
                The parent tree node.
            blocks:
                The list of text blocks which gets processed.
        """

        block: Text = blocks.pop(0)

        markup_node = self.call_processor(block)

        if markup_node is not None:
            parent.append(
                markup_node
            )


class CallInlineProcessor(InlineProcessor):
    """Processor for inline calls.

    Attributes:
        call_processor:
            The processor for handling calls.
    """

    def __init__(
            self,
            pattern,
            call_processor: CallProcessor,
            md: Markdown = None
    ):
        """Initializes the inline call processor."""
        super().__init__(pattern, md=md)

        self.call_processor = call_processor

    def handleMatch(self, m, data: Text) -> Tuple[etree.Element, int, int]:
        markup_node: etree.Element = self.call_processor(
            data[m.start(0):m.end(0)]
        )
        return markup_node, m.start(0), m.end(0)


class TemplatePostProcessor(Postprocessor):

    def run(self, text: Text) -> Text:
        """Wraps the existing markup tree in a template with meta data."""

        data = self.md.meta_data
        importer = data.get("imports", Importer(dict()))
        template = importer.template
        meta_data = importer.meta_data

        if template is not None:
            self.md.stripTopLevelTags = False
            meta_data["content"] = Markup(
                text
            )
            result = template(meta_data)

            return result

        return text


def makeExtension(**kwargs):  # pylint: disable=invalid-name
    """Function returning an instance of the extension."""

    return MetaExtension(**kwargs)


def get_parser() -> Markdown:
    """Returns a Markdown parser including the Markdown Meta Extension."""
    return Markdown(
        extensions=[
            MetaExtension()
        ]
    )


def convert(input_path: Text, output_path: Text = None):
    """Converts an input Markdown file to an output HTML file.

    Args:
        input_path:
            The path to the input Markdown file.
        output_path:
            Optional path for the HTML output.
            If not provided, path is based on input path.
    """
    parser = get_parser()
    input_path = Path(input_path)

    if not output_path:
        output_path = input_path.with_suffix(".html")
    else:
        output_path = Path(output_path)

    output_path.write_text(
        parser.convert(input_path.read_text())
    )


def convert_from_string(input_string: Text) -> Text:
    parser = get_parser()
    return parser.convert(input_string)


@click.command()
@click.argument(
    "input_path",
    required=True,
    type=click.Path(writable=True)
)
@click.option(
    "output_path",
    "--output",
    "-o",
    required=False,
    default=None,
    type=click.Path(),
    help="The optional path for the HTML output."
)
@click.option(
    "quiet",
    "--quiet",
    "-q",
    is_flag=True,
    flag_value=True,
    help="Mute all command line output."
)
@click.option(
    "verbose",
    "--verbose",
    "-v",
    is_flag=True,
    flag_value=True,
    help="Display additional command line output and file contents."
)
@click.option(
    "yes",
    "--yes",
    "-y",
    is_flag=True,
    flag_value=True,
    help="Confirm overwriting existing output file."
)
def markdown_meta_extension_parser(
        input_path: Text,
        output_path: Text = None,
        quiet: bool = False,
        verbose: bool = False,
        yes: bool = False
):
    """CLI command for converting an input Markdown file to an output HTML file.

    Args:
        input_path:
            The path to the Markdown file to get processed.
        output_path:
            The optional path for the output HTML file.
            If not provided, it is based on the input path.
        quiet:
            Flag for muting console output for this command.
        verbose:
            Flag whether to display additional information during conversion.
        yes:
            Flag whether to skip overwrite confirmation dialog with yes.
    """
    input_path = Path(input_path)

    if verbose and not quiet and input_path.suffix not in [".md", ".markdown"]:
        click.echo(f"Input file {input_path} has an unexpected extension.")

    if not output_path:
        output_path = input_path.with_suffix(".html")
        if not quiet:
            click.echo(f"Output path {output_path} is assumed.")
    else:
        output_path = Path(output_path)

    parser = get_parser()

    if verbose and not quiet:
        click.echo(
            "Markdown parser with Markdown Meta Extension "
            "successfully initialized."
        )

    input_string = input_path.read_text()

    if verbose and not quiet:
        click.echo()
        click.echo(input_string)

    output_string = parser.convert(input_path.read_text())

    if not quiet:
        # pylint: disable=no-member
        number_imports = len(parser.meta_data.get("imports", dict()))
        click.echo(f"{number_imports} callables were imported.")

        number_calls = parser.meta_data["statistics"].get("calls", 0)
        click.echo(f"{number_calls} calls were made.")

    if verbose and not quiet:
        click.echo()
        click.echo(output_string)

    write_file = True
    if output_path.exists():
        write_file = (
            yes or
            click.confirm(f"Confirm overwriting output file {output_path}.")
        )
        if write_file:
            if not quiet:
                click.echo(f"File {output_path} will be overwritten.")
        else:
            if not quiet:
                click.echo(f"Aborted writing to file {output_path}.")

    if write_file:
        output_path.write_text(output_string)
        if not quiet:
            click.echo(f"Output written to file {output_path}.")

@click.command()
@click.argument(
    "glob",
    required=True,
    type=click.STRING
)
@click.option(
    "directory",
    "--directory",
    "-d",
    required=False,
    default=None,
    type=click.Path(dir_okay=True, exists=True),
    help="The optional directory for the glob pattern."
)
def markdown_meta_extension_glob_parser(
        glob,
        directory
):
    """Parses files corresponding to glob pattern in directory.

    Args:
        glob:
            The glob pattern to match the files to get parsed.
        directory:
            The optional starting point for the glob pattern.
            Defaults to current working directory.
    """
    if not directory:
        directory = Path.cwd()
    else:
        directory = Path(directory)

    input_file_paths = list(directory.glob(glob))

    with click.progressbar(input_file_paths) as progress_bar:
        for input_file_path in progress_bar:
            convert(
                input_file_path
            )


if PELICAN:

    class MarkdownMetaExtensionPelicanReader(MarkdownReader):
        """Reader for Markdown Meta Extension files."""

        enabled = bool(MetaExtension)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            settings = self.settings["MARKDOWN"]

            if "markdown_meta_extension" not in settings["extensions"]:
                settings["extensions"].append("markdown_meta_extension")

            settings["extensions"].pop("markdown.extensions.meta")

        def _parse_metadata(self, meta):
            """Return the dict containing document metadata"""

            return NotImplementedError

        def read(self, source_path):
            """Parse content and metadata of markdown files"""

            self._source_path = source_path
            self._md = Markdown(**self.settings["MARKDOWN"])
            with pelican_open(source_path) as text:
                content = self._md.convert(text)

            if hasattr(self._md, "meta_data"):
                # pylint: disable=no-member
                metadata = self._md.meta_data.get("meta-data", dict())
            else:
                metadata = dict()

            return content, metadata



"""Importing Module.

This Module provides the Importer class, which is used for processing
the imports frontmatter of a document.

Author:
    Martin Schorfmann
Since:
    2019-12-14
Version:
    2020-02-27
"""

import collections
from pathlib import Path
from typing import Any, Callable, Dict, List, Text, Union

import yaml

from .plugins.spec import CallableWrapper


INTER_PATH_SEPARATOR = ":"  # TODO: Import instead of hardcoding


class Importer(collections.UserDict):  # pylint: disable=too-many-ancestors
    """Imports callables according to frontmatter dict.

    Imports callables according to the frontmatter dict upon instantiation.

    Attributes:
        source:
            The source dictionary containing textual or list import statements
            mapped to keys
        data:
            Internal dictionary mapping from a key to a callable wrapper
            for each key.
    """

    source: Dict[Text, Text]
    data: Dict[Text, CallableWrapper]
    meta_data: Dict[Text, Any]

    def __init__(
            self,
            source: Union[Text, Dict[Text, Union[Text, List[Text]]]]
    ):
        """Initializes Importer with the import frontmatter.

        Processes the import frontmatter and referenced external import files
        into a single dictionary for accessing the callable wrappers by key.

        Args:
            source:
                The mapping from key to import statements in form of a dict or
                a parsable YAML source string.
        """

        if isinstance(source, str):
            source = yaml.safe_load(source)

        meta_data = source.pop("meta-data", dict())

        predefined_source = self.process_predefined(
            source.pop("import", list())
        )

        predefined_source.update(source)

        self.source = predefined_source

        super().__init__(
            self.process_source(
                self.source
            )
        )

        predefined_meta_data = self.process_predefined(
            source.pop("meta-data-import", list())
        )

        predefined_meta_data.update(meta_data)

        template = predefined_meta_data.pop("template", None)

        if template:
            if not isinstance(template, list):
                template = template.split(INTER_PATH_SEPARATOR)

            self.template = CallableWrapper.import_callable(
                import_key="template",
                import_statement=template
            )
        else:
            self.template = None

        self.meta_data = predefined_meta_data

    @classmethod
    def process_source(
            cls,
            source: Dict[Text, Union[Text, List[Text]]]
    ) -> Dict[Text, Callable]:
        """Processes the source import statements.

        Args:
            source:
                The source mapping from key to import statement
                (Either string with separator chars or list).

        Returns:
            The processed imports containing callable wrappers as values.
        """

        processed_source = dict()

        for key, callable_path in source.items():

            if not isinstance(callable_path, list):
                callable_path = callable_path.split(INTER_PATH_SEPARATOR)

            processed_source[key] = CallableWrapper.import_callable(
                import_key=key,
                import_statement=callable_path,
                plugin_name=None  # TODO: Pass plugin name
            )

        return processed_source

    @classmethod
    def process_predefined(
            cls,
            predefined: List[Text]
    ) -> Dict[Text, Text]:
        """Processes external import YAML files.

        Reads in and merges predefined import YAML files.

        Args:
            predefined:
                List of file paths.

        Returns:
            Mapping from keys to import statements, where
            earlier non-unique keys got overridden by later keys of the same
            key value.
        """

        predefined_source = dict()

        for predefined_item in predefined:
            predefined_item_path = Path(predefined_item)

            predefined_source.update(
                yaml.safe_load(predefined_item_path.read_text())
            )

        return predefined_source



"""Callable Wrapper and Handler Specification.

This module specifies the wrapper around importable callables and the class
used by handler plugins.
Instances of the class provide functionality around the import and
use of certain callables according to import statements.

Author:
    Martin Schorfmann
Since:
    2020-01-26
Version:
    2020-02-27
"""

from typing import Any, Callable, Dict, List, Text, Union


class CallableWrapper:
    """Wrapper class around callables.

    Wrapper class around callables, which provides import, call and get
    functionality by using implemented instances of CallableHandler.

    Attributes:
        import_key:
            The key the callable wrapper or following callable wrappers
            will be accessible by.
        import_statement:
            The import statement list used by the callable handlers
            for importing the callable.
        plugin_name:
            An optional plugin name to give handling to the plugin of this name.
        callable_handler:
            The callable handler used to manage this callable.
        parent:
            The parent callable wrapper if this wrapper is not the first stage.
        imported_callable:
            The imported callable for use or the intermediate object for
            further importing.
    """

    import_key: Text
    import_statement: List[Text]
    plugin_name: Text
    callable_handler: "CallableHandler"
    parent: "CallableWrapper"
    imported_callable: Any

    @classmethod
    def import_callable(
            cls,
            import_key: Text,
            import_statement: List[Text],
            plugin_name: Text = None
    ) -> "CallableWrapper":
        """Imports the final callable.

        Imports the final callable through a sequence of intermediate callable
        wrappers.

        Args:
            import_key:
                The key the final callable wrapper will be accessible by.
            import_statement:
                The list of import statement parts used for the import.
            plugin_name:
                The optional name of the plugin to use for importing.

        Returns:
            The final callable wrapper of the import.
        """

        imported_callable_wrapper = None

        # pylint: disable=unused-variable
        while imported_callable_wrapper is None or (
                imported_callable_wrapper.callable_handler\
                and imported_callable_wrapper.callable_handler\
                    .has_following_handlers()
        ):
            # TODO: How to determine iterations?
            imported_callable_wrapper = CallableWrapper(
                import_key=import_key,
                import_statement=import_statement,
                plugin_name=plugin_name,
                parent=imported_callable_wrapper
            )

        if not imported_callable_wrapper.is_functional():
            return imported_callable_wrapper.parent

        return imported_callable_wrapper


    def __init__(
            self,
            import_key,
            import_statement: Text,
            plugin_name: Text = None,
            callable_handler=None,
            parent=None,
            imported_callable=None
    ):
        """Initializes a callable wrapper.

        Args:
            import_key:
                The key the callable wrapper or following callable wrappers
                will be accessible by.
            import_statement:
                The import statement list used by the callable handlers
                for importing the callable.
            plugin_name:
                An optional plugin name to give handling to the plugin of this name.
            callable_handler:
                An optional preselected callable handler.
            parent:
                The optional parent callable wrapper if this wrapper is not the first stage.
            imported_callable:
                Optionally the imported callable
        """

        self.import_key = import_key
        self.import_statement = import_statement
        self.plugin_name = plugin_name
        self.parent = parent
        self.imported_callable = imported_callable

        if not callable_handler:
            self.callable_handler = CallableHandler.assign_handler_type(
                self
            )
        else:
            self.callable_handler = callable_handler

        if self.callable_handler:
            self.imported_callable = self.callable_handler.import_callable(
                self
            )

    def parent_handler_name(self) -> Text:
        """Shortcut method for the name of the parent handler."""

        if self.parent and self.parent.callable_handler:
            return self.parent.callable_handler.name

        return None

    def parent_imported_callable(self) -> Any:
        """Shortcut method for the imported callable of the parent handler."""

        if self.parent and self.parent.imported_callable:
            return self.parent.imported_callable

        return None

    def is_functional(self) -> bool:
        """Whether the callable is functional."""

        return bool(self.callable_handler and self.imported_callable)

    def __bool__(self) -> bool:
        return self.is_functional()

    def __call__(self, arguments=None) -> Any:
        """Delegates call to callable handler and returns its result."""

        if self.is_functional():
            return self.callable_handler(self, arguments)
        else:
            try:
                return self.callable_handler(self, arguments)
            except TypeError:
                return self.imported_callable

    def __getitem__(self, key: Union[Text, List]) -> "CallableWrapper":
        """Delegates getting of an item to the handler returns item wrapper."""

        if self.is_functional():
            if isinstance(key, list):
                callable_wrapper = None

                for partial_key in key:
                    callable_wrapper = self.callable_handler.get_child(
                        self,
                        partial_key
                    )

            else:
                callable_wrapper = self.callable_handler.get_child(
                    self,
                    key
                )

            return callable_wrapper

        return None


class CallableHandler:
    """Handlers for handling importing and calling of callables.

    Attributes:
        all_handlers:
            Mapping from handler names to callable handler instances.
        initial_handlers:
            List of initally usable handlers.
        following_handlers:
            Mapping from handler keys to a lists of following handlers.
        name:
            Name of the handler instance.
        requires:
            Required handler names to come before this handler.
        check_function:
            Function used to check if this handler is applicable for
            a callable wrapper.
        import_function:
            Function used for proceeding with the callable import
            in this stage.
        call_function:
            Function used for defining call behavior of the callable wrapper.
        get_function:
            Function used for getting child elements of the callable wrapper.
    """

    all_handlers: Dict[Text, "CallableHandler"] = dict()
    initial_handlers: List["CallableHandler"] = list()
    following_handlers: Dict[Text, "CallableHandler"] = dict()

    def __init__(
            self,
            name: Text,
            requires: Union[Text, List[Text]],
            check_function: Callable[
                [CallableWrapper], bool
            ],
            import_function: Callable[
                [CallableWrapper], Any
            ] = None,
            call_function: Callable[
                [CallableWrapper, Union[List, Dict]], Any
            ] = None,
            get_function: Callable[
                [CallableWrapper, Text], CallableWrapper
            ] = None
    ):
        """Initializes a handler instance.

        Also registers this handler with its class.

        Args:
            name:
                Name of the handler instance.
            requires:
                Required handler names or single name to come
                before this handler.
            check_function:
                Function used to check if this handler is applicable for
                a callable wrapper.
            import_function:
                Function used for proceeding with the callable import
                in this stage.
            call_function:
                Function used for defining call behavior of
                the callable wrapper.
            get_function:
                Function used for getting child elements of
                the callable wrapper.
        """

        self.name = name
        self.requires = requires if isinstance(requires, list) else [requires]
        self.check_function = check_function
        self.import_function = import_function
        self.call_function = call_function
        self.get_function = get_function

        CallableHandler.register_handler(self)

    @classmethod
    def register_handler(
            cls,
            callable_handler: "CallableHandler"
    ):
        """Registers the handler instance with its class."""

        name = callable_handler.name
        cls.all_handlers[name] = callable_handler

        for requirement in callable_handler.requires:
            if requirement is None:
                cls.initial_handlers.append(callable_handler)
            else:
                if requirement not in cls.following_handlers:
                    cls.following_handlers[requirement] = list()

                cls.following_handlers[requirement].append(callable_handler)

    @classmethod
    def assign_handler_type(
            cls,
            callable_wrapper: CallableWrapper
    ) -> "CallableHandler":
        """Assigns the fitting handler instance to the callable wrapper"""
        if not callable_wrapper.parent:
            callable_handlers = cls.initial_handlers
        else:
            callable_handlers = cls.following_handlers.get(
                callable_wrapper.parent_handler_name(),
                list()
            )

        for callable_handler in callable_handlers:
            if callable_handler.check(
                    callable_wrapper=callable_wrapper
            ):
                return callable_handler

        return None

    def has_following_handlers(self) -> bool:
        """Whether this handler has following handlers."""

        return bool(
            self.following_handlers.get(self.name)
        )

    def check(
            self,
            callable_wrapper
    ) -> bool:
        """Wrapper function around check function."""

        return bool(self.check_function(callable_wrapper))

    def import_callable(
            self,
            callable_wrapper
    ) -> Any:
        """Imports callable using this handler or returns parent callable."""
        if self.import_function:
            imported_callable = self.import_function(
                callable_wrapper
            )
        elif callable_wrapper.parent:
            imported_callable = callable_wrapper.parent_imported_callable()
        else:
            imported_callable = None

        return imported_callable


    def __call__(
            self,
            callable_wrapper: CallableWrapper,
            arguments: Union[List[Any], Dict[Text, Any]]
    ) -> Any:
        """Handles calls for callable wrapper.

        Defines some default behavior if a custom call function is absent.

        Args:
            callable_wrapper:
                The callable wrapper whose callable will be called.
            arguments:
                Ordered arguments or keyworded arguments
                (Includes no arguments).

        Returns:
            The result of the call to the callable wrapper's callable.
        """

        if not self.call_function:
            try:
                if not arguments:
                    returned_result = callable_wrapper.imported_callable()
                elif isinstance(arguments, dict):
                    returned_result = callable_wrapper.imported_callable(
                        **arguments
                    )
                elif isinstance(arguments, list):
                    returned_result = callable_wrapper.imported_callable(
                        *arguments
                    )
                else:
                    raise TypeError(
                        "Type of call arguments '{arguments}' not considered.".format(
                            arguments=arguments
                        )
                    )
            except TypeError:  # TODO: Error handling
                returned_result = callable_wrapper.imported_callable
        else:
            returned_result = self.call_function(
                callable_wrapper=callable_wrapper,
                arguments=arguments
            )

        return returned_result

    def get_child(
            self,
            callable_wrapper: CallableWrapper,
            key: Text
    ) -> CallableWrapper:
        """Returns the child element of the callable wrapper.

        Defines some default behavior in case of absence of
        a custom get function.

        Args:
            callable_wrapper:
                The callable wrapper whose callable's child elements gets
                accessed.
            key:
                The key for accessing the child element.

        Returns:
            The accessed child element wrapped in a callable wrapper.
        """

        if not self.get_function:
            if hasattr(callable_wrapper.imported_callable, "__getitem__"):
                child_item = callable_wrapper.imported_callable[key]
            else:
                child_item = getattr(
                    callable_wrapper.imported_callable,
                    key
                )
        else:
            child_item = self.get_function(
                callable_wrapper=callable_wrapper,
                key=key
            )

        if not isinstance(child_item, CallableWrapper):
            child_item = CallableWrapper(
                import_key=callable_wrapper.import_key,
                import_statement=callable_wrapper.import_statement + [key],
                plugin_name=callable_wrapper.plugin_name,
                callable_handler=None,
                parent=callable_wrapper,
                imported_callable=child_item
            )

        return child_item



"""Package Handler Plugin.

This plugin module provides the handler for handling import statements
for (installed) Python packages.

Author:
    Martin Schorfmann
Since:
    2020-01-28
Version:
    2020-02-25
"""

import importlib
import types

from markdown_meta_extension.plugins.spec import (
    CallableWrapper,
    CallableHandler
)

def check_package_path(
        callable_wrapper: CallableWrapper
) -> bool:
    """Checks whether the callable wrapper imports a Python package.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        Whether the callable wrapper imports a Python package.
    """

    try:
        return bool(
            importlib.util.find_spec(
                callable_wrapper.import_statement[0]
            )
        )
    except ValueError:
        return False

def import_package(
        callable_wrapper: CallableWrapper
) -> types.ModuleType:
    """Imports the callable wrapper's Python package.

    Args:
        callable_wrapper:
            The callable wrapper to get imported.

    Returns:
        The imported Python module.
    """
    package_path = callable_wrapper.import_statement[0]

    try:
        imported_module = importlib.import_module(package_path)
    except ImportError:  # TODO: Error handling
        imported_module = None

    return imported_module


PACKAGE_HANDLER = CallableHandler(
    name="package",
    requires=None,
    check_function=check_package_path,
    import_function=import_package
)

def get_callable_plugin() -> CallableHandler:
    """Returns the callable handler for importing Python packages."""
    return PACKAGE_HANDLER



"""Setup for installing the package.

This module also provides configuration options and meta data for the package.

Author:
    Martin Schorfmann
Since:
    2019-12-11
Version:
    2020-03-10
"""

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta data
NAME = "markdown-meta-extension"
DESCRIPTION = (
    "Markdown Meta Extension for using "
    "Python functions and classes from within documents."
)
URL = "https://gitlab.com/markdown-meta-extension/markdown-meta-extension"
EMAIL = "schorfma@uni-bremen.de"
AUTHOR = "Martin Schorfmann"
REQUIRES_PYTHON = ">=3.7.0"
VERSION = "0.5.2"

LICENSE = "MIT"
LICENSE_FILE = "./LICENSE.md"

# Requires packages for execution
REQUIRED = [
    "Click",
    "jinja2",
    "Markdown",
    "MarkupSafe",
    "pathvalidate",
    "PyYAML"
]

TESTS_REQUIRED = [
    "beautifulsoup4",
    "lxml",
    "Markdown",
    "parse"
]

# Optional packages
EXTRAS = {
    "pelican": [
        "pelican"
    ]
}

ENTRY_POINTS = {
    "markdown.extensions": [
        "markdown_meta_extension = markdown_meta_extension.__init__"
    ],
    "console_scripts": [
        "markdown-meta-extension = markdown_meta_extension:markdown_meta_extension_parser",
        "markdown-meta-extension-glob = markdown_meta_extension:markdown_meta_extension_glob_parser"
    ]
}

# https://github.com/navdeep-G/setup.py
here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if "README.md" is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as file:
        long_description = "\n" + file.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package"s __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as file:
        exec(file.read(), about)
else:
    about["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(
        exclude=[
            "tests",
            "*.tests",
            "*.tests.*",
            "tests.*"
        ]
    ),
    # If your package is a single module, use this instead of "packages":
    # py_modules=["mypackage"],

    entry_points=ENTRY_POINTS,
    install_requires=REQUIRED,
    tests_require=TESTS_REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license=LICENSE,
    license_file=LICENSE_FILE,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Text Processing :: Markup",
        "License :: OSI Approved :: MIT License"
    ],
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
    },
)
